module audio.audio;
import derelict.sdl.sdl;
import audio.filter;
import audio.player;
import std.stdio;
import std.c.stdlib;
import std.c.string;
alias writefln w;

const MIXBUF_MUL = 2;
int buffersize, bufferused; // in samples
short* mixbuf;
SDL_AudioSpec audiospec;
int framerate = 50, cbinterval;
int multiplier;
int accurate = 1;
int cbcounter = 0;

extern(C) {
	void (*callback)();

	int audio_init(int b, int f, int fr, void (*cb)())
	{
		SDL_AudioSpec requested;

		framerate = fr;
		buffersize = b;
		if(b < f / framerate) {
			fprintf(stderr,"Minimum buffer size is %d bytes.\n", f / framerate);
			return -1;
		}
		
		requested.freq = f;
		requested.format = AUDIO_S16LSB;
		requested.channels = 1;
		requested.samples = buffersize;

		if(accurate == 1)
			requested.callback = &audio_callback_2;
		else 
			requested.callback = &audio_callback;
		
		requested.userdata = null;
		callback = cb;
		if(SDL_OpenAudio(&requested, &audiospec) < 0) {
			fprintf(stderr,"Could not open audio: %s\n", SDL_GetError());
			return -1;
		}
		if(audiospec.format != AUDIO_S16LSB) {
			fprintf(stderr,"Incorrect audio format obtained.\n");
			return -1;
		}
		if(buffersize != audiospec.samples) {
			fprintf(stderr,"Warning: obtained buffer size differs (tried %d, got %d).\n",
					buffersize, audiospec.samples);
			buffersize = audiospec.samples;
		}
		mixbuf = cast(short *)malloc(buffersize * short.sizeof * MIXBUF_MUL);
		setMultiplier(1);
		return 0;
	}

	void audio_close() {
		SDL_CloseAudio();
		if(mixbuf) free(mixbuf);
		sid_close();
	}
	
	void reset() {
		bufferused = cbcounter = 0;
	}

	void setMultiplier(int m) {
		if(m < 1 || m > 16) return;
		if(m == multiplier) return;
		SDL_LockAudio();
		multiplier = m;
		framerate = 50 * m;
		cbinterval = audiospec.freq / framerate;
		reset();
		SDL_UnlockAudio();
	}


	void audio_callback(void *data, ubyte* stream, int len) {
		int samples = len / short.sizeof;
		int total = 0,todo = 0,t = 0;
		int steps;
		if(!isPlaying) return;
		while(total < samples) {
			todo = samples - total;
			assert(todo >= 0);
			if(cbcounter + todo >= cbinterval) {
				int c = cbinterval - cbcounter;
				assert(c > 0);
				t = sid_fillbuffer(cast(short*)stream + total,
								   c, 0);
				(*callback)();
				cbcounter -= cbinterval;
				assert(cbcounter + todo >= 0);
			}
			else {
				assert(total + todo <= samples);
				t = sid_fillbuffer(cast(short*)stream + total,
								   todo, 0);
				
			}
			total += t;
			cbcounter += t;
			steps++;
		}
		assert(total == samples);
	}

	void audio_callback_2(void *data, ubyte* stream, int len) {
		int samples = len / short.sizeof;
		int i,j,t;

		if(!isPlaying) return;
		

		while((bufferused + cbinterval) <= buffersize * MIXBUF_MUL) {
			t = sid_fillbuffer(mixbuf+bufferused, cbinterval, cyclesPerFrame);
			bufferused += t;
			(*callback)();
		}
		memcpy(stream, cast(ubyte*)mixbuf, len);
		bufferused -= samples;
		if(bufferused < 0) {
			fprintf(stderr, "Audio buffer underrun!\n");
			bufferused = 0;
		}

		short* pi, po;
		for(i = 0, pi = mixbuf, po = mixbuf + samples;
			i < buffersize * MIXBUF_MUL - samples; i++) {
			*(pi++) = *(po++);
		}
	}
}

extern(C) {
	extern char[0x19] sidreg;
	extern int residdelay, np21;
	int sid_init(int, Filterparams*, int, int, int, int, int);
	int sid_fillbuffer(short *, int, int);
	int sid_close();
}
