

// TODO
// swingcat.ct does not pack properly (only 2nd voice plays)

module ct.pack;
import ct.base;
import com.cpu;
import std.stdio;
import std.string;


alias writefln w;

const ubyte[] SIDHEADER = [
  0x50, 0x53, 0x49, 0x44, 0x00, 0x02, 0x00, 0x7c, 0x00, 0x00, 0x10, 0x00,
  0x10, 0x03, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x53, 0x77,
  0x61, 0x6d, 0x70, 0x20, 0x50, 0x6f, 0x6f, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x54, 0x68, 0x6f, 0x6d, 0x61, 0x73,
  0x20, 0x4d, 0x6f, 0x67, 0x65, 0x6e, 0x73, 0x65, 0x6e, 0x20, 0x28, 0x44,
  0x52, 0x41, 0x58, 0x29, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x32, 0x30, 0x30, 0x34, 0x20, 0x4d, 0x61, 0x6e, 0x69, 0x61,
  0x63, 0x73, 0x20, 0x6f, 0x66, 0x20, 0x4e, 0x6f, 0x69, 0x73, 0x65, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x14,
  0x00, 0x00, 0x00, 0x00 
];

enum {
	PSID_LOAD_ADDR_OFFSET = 0x08,
	PSID_INIT_OFFSET = 0x0a,
	PSID_PLAY_OFFSET = 0x0c,
	PSID_TITLE_OFFSET = 0x16,
	PSID_FLAGS_OFFSET = 0x76,
	PSID_NUM_SONGS = 0x0e,
	PSID_START_SONG = 0x10,
	PSID_SPEED_OFFSET = 0x15,
	CIA_OFFSET = 0x09,
	DIV_COUNTER = 0x1b,
	PAL_CLOCK = 0x4cc7
}

static const auto SIDDriver = cast(ubyte[])import("custplay.bin");

private class DataRelocator : CPU {
	address exec;
	address relocUpto;
	address relocFrom, relocTo;

	this(ubyte[] mem) {
		super(mem);
	}

	void relocate(address ex, address from, address to, address upto) {
		exec = ex; relocFrom = from; relocTo = to; relocUpto = upto;
		reset();
		execute(ex, false);
	}

	override void handleBreak() {
		return;
	}

	protected void rewriteAddress(ubyte[] chunk) {
		address addr = toAddress(chunk[1..$]);
		if(addr < relocFrom || (relocUpto > 0 && addr >= relocUpto)) return;
		if(chunk[2] >= 0xd0 && chunk[2] < 0xe0) return;
		address baseOld = relocFrom, baseNew = relocTo;
		int oldHi = baseOld; 
		int newBaseHi = baseNew;
		int delta = newBaseHi - oldHi;
		int newaddr = chunk[2] * 256 + delta + chunk[1];
		
		chunk[1] = lowbyte(newaddr);
		chunk[2] = highbyte(newaddr);
	}

	override ushort fetch(State st) {
		ushort arg = st.arg;
		switch(st.addrmode) {
		case Am.ABSOLUTE:
			rewriteAddress(st.chunk);
			return memory[arg];
	    case Am.ABSOLUTE_X:
			rewriteAddress(st.chunk);
			return memory[(arg + regs.x) & 65535];
	    case Am.ZY, Am.ABSOLUTE_Y:
			rewriteAddress(st.chunk);
			return memory[(arg + regs.y) & 65535];
		case Am.RELATIVE:
			return regs.pc + 2;
		default:
			return super.fetch(st);
		}
	}	

	override void executeOp(inout State stat) {
		switch(stat.opcode) {
		case Op.JMP:
			regs.pc += 3;
			break;
		case Op.RTS:
			regs.pc += 1;
			break;
		default:
			super.executeOp(stat);
		}
	}
}

					   
private void explain(string s) {
	if(beVerbose) writefln(s);
}

private Song sng;
private DataRelocator relocator;
private bool beVerbose = false;

ubyte[] pack(Song insong, address relocTo, bool verbose) {
	sng = insong;
	memspace = sng.memspace.dup;
	if(relocator is null)
		relocator = new DataRelocator(memspace);

	beVerbose = verbose;
	ubyte[] data = packSongdata(relocTo);
	data = cast(ubyte[])toArr(relocTo) ~ data;
	
	return data;
}



/** PSID TODO:
	- pack instrument table
	- handle NTSC files
	- handle relocating of multispeed tunes
	- handle speeds & voiceflags set from commandline
	- handle SID filter revision
   maybe:
    - optimize freqtable
*/
ubyte[] packToSid(Song insong, address relocTo, int defaultSubtune, bool verbose) {
	ubyte[] data = pack(insong, relocTo, verbose);
	if(defaultSubtune >= sng.subtunes.numOf)
		throw new Error(format("This song only has ", sng.subtunes.numOf, " subtunes."));
	if(sng.multiplier > 1) {
		ubyte[] custplay = SIDDriver.dup; 

		int clock = PAL_CLOCK / sng.multiplier;
		
		custplay[CIA_OFFSET] = highbyte(clock);
		custplay[CIA_OFFSET + 2] = lowbyte(clock);
		custplay[DIV_COUNTER] = sng.multiplier - 1;
		data = custplay ~ data[2 .. $];
	}
	data = SIDHEADER ~ data;
	void outstr(char[] s, int offset) {
		data[offset .. offset + s.length] = cast(ubyte[])s;
	}
	data[PSID_TITLE_OFFSET .. PSID_TITLE_OFFSET + 0x20] = '\0';
	data[PSID_TITLE_OFFSET + 0x20 .. PSID_TITLE_OFFSET + 0x40] = '\0';
	data[PSID_TITLE_OFFSET + 0x40 .. PSID_TITLE_OFFSET + 0x60] = '\0';
	outstr(stripr(sng.title),PSID_TITLE_OFFSET);
	outstr(stripr(sng.author),PSID_TITLE_OFFSET + 0x20);
	outstr(stripr(sng.release),PSID_TITLE_OFFSET + 0x40);
	data[PSID_NUM_SONGS + 1] = sng.subtunes.numOf();
	data[PSID_START_SONG + 1] = defaultSubtune;
	if(sng.multiplier > 1) {
		if(relocTo != 0x1000) throw new Error("Relocating multispeed tunes not yet supported.");
		data[PSID_SPEED_OFFSET] |= 1;
		data[PSID_INIT_OFFSET .. PSID_INIT_OFFSET + 2] = [ 0x0f, 0xdc];
		data[PSID_PLAY_OFFSET .. PSID_PLAY_OFFSET + 2] = [ 0x0f, 0xdf];
	}
	else {
		data[PSID_INIT_OFFSET .. PSID_INIT_OFFSET + 2] = [ relocTo >> 8, relocTo & 255 ];
		data[PSID_PLAY_OFFSET .. PSID_PLAY_OFFSET + 2] = [ (relocTo + 3) >> 8, (relocTo + 3) & 255 ];
	}
	data[PSID_FLAGS_OFFSET + 1] 
		= 0x04 /+ PAL +/ | (sng.sidModel ? 0x20 : 0x10);

	return data;
}

private:
	ubyte[] memspace;
	struct Songset {
		address[3] voices;
		ubyte speed, mask;
	}
	ubyte[][][] packedTracks;
	address[][] trackpointers;
	ubyte[][] packedSeqs;
	ubyte[] packedData;
	Songset[] songsets;
	address[] seqpointers;
	ubyte[] playerCode;
	address dataStart, dataOffset;
	address inputStart, inputOffset;
	address codeStart;
	int numOfSeqs;

/*
  NOTE: the packer assumes that the first jmp in the jumptable points
  to the lowest address in the player code. This is done so
  that players can place some data between the jump table and
  the player code. Placing data in the middle of the player code
  will crash the packer.

  FIX: the relocator must relocate the jump table separately.

  returns the packed song.
*/
ubyte[] packSongdata(address reloc) {
	numOfSeqs = sng.numOfSeqs();

	void output(ubyte[] data) {
		memspace[dataOffset .. dataOffset + data.length] = data;
		dataOffset += data.length;
	}

	// ------------------------------------------------------------
	// initialize packer
	
	// seek code end signature $fc, $3c
	int endsig = hunt(memspace[0x1000 .. 0x2000], [0xfc, 0x3c]);
	if(endsig < 0) {
		throw new Error("Could not determine player code end address."); 
	}
					  
	dataStart = 0x1000 + endsig;
	dataOffset = dataStart;
	playerCode = memspace[0x1000 .. dataStart];
	codeStart = toAddress(memspace[0x1001 .. 0x1003]);
	inputOffset = sng.offsets[Offsets.Songsets];

	// ------------------------------------------------------------
	// pack seqs & tracks

	packedSeqs.length = numOfSeqs; 
	foreach(idx, seq; sng.seqs[0..numOfSeqs]) {
		packedSeqs[idx] = seq.compact();
	}

	//sng.initSubtune(0);
	sng.subtunes.activate(0);

	trackpointers.length = sng.subtunes.numOf();
	packedTracks = sng.subtunes.compact();
	int trkOffset;
	foreach(i, ref subtune; packedTracks) {
		trackpointers[i].length = 3;
		foreach(j, ref voice; subtune) {
			trackpointers[i][j] = trkOffset;
			trkOffset += voice.length;
		}
	}

	// ------------------------------------------------------------
	// generate seqpointers (data offset will be added later)

	seqpointers.length = numOfSeqs; 
	{
		int offset;
		foreach(idx, seq; packedSeqs) {
			seqpointers[idx] = offset;
			offset += seq.length;
		}
	}

	// ------------------------------------------------------------
	// generate songsets
		
	//explain(format(sng.subtunes.numOf, " subtunes"));
	songsets.length = sng.subtunes.numOf;

	address trackDataStart = dataStart + songsets.length * Songset.sizeof;

	int trackDataOffset = trackDataStart;
	foreach(sidx, ref songset; songsets) {
		ubyte[][] subtuneTracks = packedTracks[sidx];
		foreach(vidx, ref voiceAddr; songset.voices) {
			voiceAddr = trackDataOffset;
			trackDataOffset += subtuneTracks[vidx].length;
		}
		songset.speed = sng.songspeeds[sidx];
		songset.mask = 7;
	}

	// ------------------------------------------------------------
	// output songsets & tracks

	foreach(songset; songsets) {
		foreach(idx, voiceAddr; songset.voices) {
			output(addr2arr(voiceAddr + reloc - 0x1000));
		}
		output([songset.speed, songset.mask]);
	}
	foreach(subtune; packedTracks) {
		foreach(voice; subtune) {
			output(voice);
		}
	}

	inputOffset += sng.tSongsets.length;
	relocator.relocate(codeStart, sng.offsets[Offsets.Songsets], dataStart,
					   inputOffset);


	// ------------------------------------------------------------
	// pack & output instrument table
	
	int maxInsno;
	sng.seqIterator((Sequence s, Element e) { 
			int insval = e.instr.value;
			if(insval > 0x2f) return;
			if(insval > maxInsno) maxInsno = insval; });
	ubyte[] packedInstable;

	maxInsno += 1;

	packedInstable.length = maxInsno * 8;

	for(int j = 0; j < maxInsno; j++) {
		for(int i = 0; i < 8; i++) {
			packedInstable[i * maxInsno + j] = sng.tInstr.data[i * 48 + j];
		}
	}
	for(int i = 0; i < 8; i++) {
		relocator.relocate(codeStart, sng.tInstr.offset + 48 * i, dataOffset + maxInsno * i, sng.tInstr.offset + 48 * i + 1);
	}
	output(packedInstable.dup);
//	w(packedInstable);

	// ------------------------------------------------------------
	// pack & output cmdtable

	int maxCmdno;
	sng.seqIterator((Sequence s, Element e) { 
			int cmdval = e.cmd.value;
			if(cmdval > 63) return;
			if(cmdval > maxCmdno) maxCmdno = cmdval; });
	ubyte[] packedCmdtable;
	maxCmdno += 1;
	packedCmdtable.length = maxCmdno * 3;
	for(int j = 0; j < maxCmdno; j++) {
		for(int i = 0; i < 3; i++) {
			packedCmdtable[i * maxCmdno + j] = sng.tSuper.data[i * 64 + j];
		}
	}
	for(int i = 0; i < 3; i++) {
		relocator.relocate(codeStart, sng.tSuper.offset + 64 * i, dataOffset + maxCmdno * i, sng.tSuper.offset + 64 * i + 1);
	}
	output(packedCmdtable.dup);

	// ------------------------------------------------------------
	// pack tables

	inputOffset = sng.offsets[Offsets.Arp1];
	with(sng) {
		foreach(idx, table; [tPulse, tFilter, tWave1, tWave2, tChord, tChordIndex]) {
			inputOffset = table.offset;
			relocator.relocate(codeStart, inputOffset, dataOffset, inputOffset + table.length);
			output(table.data[0..table.size].dup);
			//inputOffset += table.length;
		}
	}

	// ------------------------------------------------------------
	// output seqpointers
	address seqdataStart = dataOffset + seqpointers.length * 2;
	address seqptraddrlo = dataOffset;
	address seqptraddrhi = dataOffset + seqpointers.length;

	foreach(idx, ptr; seqpointers) {
		ptr += seqdataStart + reloc - 0x1000;
		memspace[seqptraddrlo] = lowbyte(ptr);
		memspace[seqptraddrhi] = highbyte(ptr);
		
		seqptraddrlo++;
		seqptraddrhi++;
	}

	// relocate seqlos
	inputOffset = sng.offsets[Offsets.SeqLO];
	assert(inputOffset == sng.offsets[Offsets.SeqLO]);
	relocator.relocate(codeStart, sng.offsets[Offsets.SeqLO], dataOffset,
					   inputOffset + 128);

	dataOffset += seqpointers.length;

	// relocate seqhis
	inputOffset = sng.offsets[Offsets.SeqHI];
	assert(inputOffset == sng.offsets[Offsets.SeqHI]);
	relocator.relocate(codeStart, sng.offsets[Offsets.SeqHI], dataOffset,
					   inputOffset + 128);

	dataOffset += seqpointers.length;

	assert(dataOffset == seqdataStart);

	inputOffset = sng.offsets[Offsets.S00];
	// ------------------------------------------------------------
	// output seqs

	foreach(seq; packedSeqs) {
		output(seq);
		inputOffset += 256; // each full seq
	}

	// ------------------------------------------------------------
	// finalize

	if(sng.ver >= 6) {
		memspace[sng.offsets[Offsets.Editorflag]] = 1;
	}

	// ------------------------------------------------------------
	// relocate 
	{
		// relocate code
		int dataEnd = dataOffset;
		relocator.relocate(0x1000 /+ code start +/, 0x1000, reloc, dataEnd);

	}
	explain(format("Packed data length is ", dataOffset - 0x1000, " bytes."));
	return cast(ubyte[])memspace[0x1000 .. dataOffset];
}

int hunt(ubyte[] arr, ubyte[] cmp) {
	for(int i = 0; i < arr.length - 1; i++) {
		if(arr[i .. i + cmp.length] == cmp)
			return i;
	}
	return -1;
}

