/*
 * Copyright (c) 2004-2008 Derelict Developers
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * * Neither the names 'Derelict', 'DerelictSDL', nor the names of its contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
module derelict.sdl.macinit.NSApplication;

version(OSX) version = darwin;

version (darwin):

import derelict.sdl.macinit.ID;
import derelict.sdl.macinit.NSMenu;
import derelict.sdl.macinit.NSObject;
import derelict.sdl.macinit.runtime;
import derelict.sdl.macinit.selectors;
import derelict.sdl.macinit.string;
import derelict.util.loader;

package:

const NSApplication NSApp;

int function(int argc, /*const*/ char *argv[]) NSApplicationMain;

static this ()
{
    SharedLib cocoa = Derelict_LoadSharedLib("Cocoa.framework/Cocoa");

    NSApp = NSApplication.sharedApplication;
    bindFunc(NSApplicationMain)("NSApplicationMain", cocoa);
}

class NSApplication : NSObject
{
    this ()
    {
        id_ = null;
    }

    this (id id_)
    {
        this.id_ = id_;
    }

    static NSApplication alloc ()
    {
        id result = objc_msgSend(cast(id)class_, sel_alloc);
        return result ? new NSApplication(result) : null;
    }

    static Class class_ ()
    {
        return cast(Class) objc_getClass!(this.stringof);
    }

    NSApplication init ()
    {
        id result = objc_msgSend(this.id_, sel_init);
        return result ? this : null;
    }

    public static NSApplication sharedApplication ()
    {
        id result = objc_msgSend(class_NSApplication, sel_sharedApplication);
        return result ? new NSApplication(result) : null;
    }

    NSMenu mainMenu ()
    {
        id result = objc_msgSend(this.id_, sel_mainMenu);
        return result ? new NSMenu(result) : null;
    }

    void setAppleMenu (NSMenu menu)
    {
        objc_msgSend(this.id_, sel_setAppleMenu, menu ? menu.id_ : null);
    }

    void setWindowsMenu (NSMenu menu)
    {
        objc_msgSend(this.id_, sel_setWindowsMenu, menu ? menu.id_ : null);
    }

    void setMainMenu (NSMenu menu)
    {
        objc_msgSend(this.id_, sel_setMainMenu, menu ? menu.id_ : null);
    }

    void setDelegate (ID object)
    {
        objc_msgSend(this.id_, sel_setDelegate, object ? object.id_ : null);
    }

    void run ()
    {
        objc_msgSend(this.id_, sel_run);
    }

    void stop (ID sender)
    {
        objc_msgSend(this.id_, sel_stop, sender ? sender.id_ : null);
    }
}